from __future__ import annotations

import json
from pathlib import Path
from lazyops.types import BaseSettings
from typing import Optional, Dict, Any, List
from .types import SlackContext


lib_path = Path(__file__).parent
context_path = lib_path.joinpath('slack_context.json')


class SlackSettings(BaseSettings):
    """
    Slack Settings
    """

    webhook_url: Optional[str] = None
    bot_token: Optional[str] = None
    app_token: Optional[str] = None

    default_user: Optional[str] = None
    default_channel: Optional[str] = None
    cache_context: Optional[bool] = True

    class Config:
        env_prefix = 'SLACK_'
        case_sensitive = False

    def load_context(self) -> SlackContext:
        """
        Loads the Slack Context
        """
        if not context_path.exists(): return SlackContext()
        return SlackContext(
            **json.loads(context_path.read_text())
        )
    
    def save_context(self, context: SlackContext) -> None:
        """
        Saves the Slack Context
        """
        context_path.write_text(json.dumps(context.model_dump(), indent = 4))



from __future__ import annotations

import gzip
from .base import BaseCompression, logger
from typing import Optional

class GzipCompression(BaseCompression):
    name: str = "gzip"
    compression_level: Optional[int] = 9

    def validate_compression_level(self):
        """
        Validates the compression level
        """
        assert self.compression_level in range(10), "Compression level must be between 0 and 9"


    def compress(self, data: bytes, level: Optional[int] = None, **kwargs) -> bytes:
        """
        Compresses the data
        """
        if level is None: level = self.compression_level
        _kwargs = self._compression_kwargs.copy()
        if kwargs: _kwargs.update(kwargs)
        return gzip.compress(data, compresslevel= level, **_kwargs)

    def decompress(self, data: bytes, **kwargs) -> bytes:
        """
        Decompresses the data
        """
        return gzip.decompress(data)
    
    




import os
import pathlib

from typing import Any, Type, Tuple, Dict, List, Union, Optional, TYPE_CHECKING
from pydantic import Field, validator
from pydantic.networks import AnyUrl
from lazyops.types.formatting import to_camel_case, to_snake_case, to_graphql_format
from lazyops.types.classprops import classproperty, lazyproperty
from lazyops.utils.serialization import Json

"""
WIP Migration to pydantic v2
"""

# For backwards compatibility
try:
    from pydantic_settings import BaseSettings as _BaseSettings
    from pydantic.v1 import BaseModel as _BaseModel
    PYDANTIC_V2 = True
except ImportError:
    from pydantic import BaseSettings as _BaseSettings
    from pydantic import BaseModel as _BaseModel
    PYDANTIC_V2 = False

# if pydantic_version.startswith('2.0'):
#     from pydantic_settings import BaseSettings as _BaseSettings
#     PYDANTIC_V2 = True
# else:
#     from pydantic import BaseSettings as _BaseSettings
#     PYDANTIC_V2 = False

if TYPE_CHECKING:
    from pydantic.main import IncEx
    from pydantic_settings import BaseSettings as _BaseSettings

class BaseSettings(_BaseSettings):

    class Config:
        env_prefix: str = ""
        case_sensitive: bool = False

    # Maintain backwards compatibility
    if PYDANTIC_V2:
        def dict(  # noqa: D102
            self,
            *,
            include: 'IncEx' = None,
            exclude: 'IncEx' = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
        ) -> Dict[str, Any]:  # noqa UP006
            """
            Model to dict
            """
            return self.model_dump(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )

    def update_config(self, **kwargs):
        """
        Update the config for the other settings
        """
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            elif isinstance(getattr(self, k), BaseSettings):
                val = getattr(self, k)
                if hasattr(val, 'update_config'):
                    val.update_config(**v)
                else:
                    val = val.__class__(**v)
                setattr(self, k, val)
            else: 
                setattr(self, k, v)

    def set_env(self):
        """
        Update the Env variables for the current session
        """
        data = self.dict(exclude_none=True)
        for k, v in data.items():
            if isinstance(v, BaseSettings):
                v.set_env()
            else:
                os.environ[self.Config.env_prefix + k.upper()] = str(v)
    
    


class BaseModel(_BaseModel):
    class Config:
        extra = 'allow'
        arbitrary_types_allowed = True
        alias_generator = to_snake_case
        
    def get(self, name, default: Any = None):
        return getattr(self, name, default)
    
    def update(self, **kwargs):
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            setattr(self, k, v)

    if PYDANTIC_V2:
        def parse_obj(self, obj: Any, **kwargs) -> 'BaseModel':
            """
            Parse an object into a model
            """
            return self.model_validate(obj, **kwargs)

    @classmethod
    def create_one(
        cls,
        **kwargs
    ) -> Tuple[Type['BaseModel'], Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        if PYDANTIC_V2:
            resource_fields = [field.name for field in cls.model_fields.values()]
        else:
            resource_fields = [field.name for field in cls.__fields__.values()]
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields}
        if PYDANTIC_V2:
            resource_obj = cls.model_validate(resource_kwargs)
        else:
            resource_obj = cls.parse_obj(resource_kwargs)
        return resource_obj, return_kwargs
    
    @classmethod
    def create_many(cls, data: List[Dict]) -> List['BaseModel']:
        """
        Create a list of resources from a list of dicts
        """
        if PYDANTIC_V2:
            return [cls.model_validate(**d) for d in data]
        else:
            return [cls.parse_obj(d) for d in data]
    
    if not PYDANTIC_V2:
        @classproperty
        def _model_fields(cls) -> List[str]:
            return [field.name for field in cls.__fields__.values()]
        
    def get_model_fields(self) -> List[str]:
        """
        Get the model fields
        """
        if PYDANTIC_V2:
            return [field.name for field in self.model_fields.values()]

        return [field.name for field in self.__fields__.values()]

    def replace(self, obj: Type['BaseModel']):
        """Replace current attributes with `obj` attributes."""
        for field in obj.get_model_fields():
            if hasattr(self, field):
                setattr(self, field, getattr(obj, field))
    
    # Maintain backwards compatibility
    if PYDANTIC_V2:
        def dict(  # noqa: D102
            self,
            *,
            include: 'IncEx' = None,
            exclude: 'IncEx' = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
        ) -> Dict[str, Any]:  # noqa UP006
            return self.model_dump(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )

    def graphql(
        self,
        *,
        include: Optional[Any] = None,
        exclude: Optional[Any] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Union[Dict, Any]:
        """
        Return the data in `graphql` format
        """
        data = self.dict(
            include = include, exclude = exclude, \
            by_alias = by_alias, skip_defaults = skip_defaults, \
            exclude_unset = exclude_unset, exclude_defaults = exclude_defaults, \
            exclude_none = exclude_none
        )
        return to_graphql_format(data)



class Schema(BaseModel):

    class Config:
        extra = 'allow'
        arbitrary_types_allowed = True
        json_dumps = Json.dumps
        # alias_generator = to_camel_case

    # @lazyproperty
    # def schema_fields(self) -> List[str]:
    #     return [field.name for field in self.__fields__.values()]

    def get(self, name, default: Any = None):
        return getattr(self, name, default)


class Constant(tuple):
    "Pretty display of immutable constant."

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return f'{self[0]}'


class UrlModel(BaseModel):
    """
    Parse a URL into its components
    """
    url: AnyUrl
    scheme: Optional[str] = 'http'
    adapter: Optional[str] = None

    @lazyproperty
    def db(self):
        return self.url.path[1:]
    
    @lazyproperty
    def path(self):
        return self.url.path

    @lazyproperty
    def host(self):
        return self.url.host

    @lazyproperty
    def port(self):
        return self.url.port

    @lazyproperty
    def user(self):
        return self.url.user or None
    
    @lazyproperty
    def password(self):
        return self.url.password or None
    
    @lazyproperty
    def api_key(self):
        return self.url.user or None
    
    @lazyproperty
    def token(self):
        return self.url.password or None

    @lazyproperty
    def server_uri(self):
        """
        Returns the bare connection uri without the database/path/auth
        """
        return self.get_uri(include_auth = False, include_db = False, include_path = False)
    
    @lazyproperty
    def base_uri(self):
        """
        Returns the connection uri without the database/path
        """
        return self.get_uri(include_path = False, include_db = False)
    
    @lazyproperty
    def safe_uri(self):
        """
        Returns the connection uri without the password
        """
        return self.get_uri(include_auth = True, include_db = True, include_path = True, include_adapter = True, _safe = True)

    @lazyproperty
    def connection_uri(self):
        """
        Returns the connection uri with the database/path
        """
        return self.get_uri(include_path = True, include_db = True, include_auth = True, include_adapter = True, _safe = False)

    def get_uri(
        self, 
        scheme: Optional[str] = None, 
        db: Optional[str] = None,
        path: Optional[str] = None,
        include_path: Optional[bool] = True,
        include_db: Optional[bool] = True,
        include_auth: Optional[bool] = True,
        include_adapter: Optional[bool] = True,
        _safe: Optional[bool] = False,
        **kwargs,
    ):
        scheme = scheme or self.scheme
        if include_adapter and self.adapter and self.adapter not in scheme:
            scheme = f'{self.adapter}+{scheme}'
        uri = f'{scheme}://'
        if include_auth and self.user:
            uri += '*****' if _safe else f'{self.user}'
            if self.password:
                uri += ':*****' if _safe else f':{self.password}'
            uri += '@'
        uri += f'{self.host}'
        if self.port:
            uri += f':{self.port}'
        if 'http' in scheme and include_path:
            path = path or self.path
            if path: uri += f'/{path}'
        elif include_db:
            db = db or self.db
            if db: uri += f'/{db}'
        return uri

    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'http',
        adapter: Optional[str] = None,
        **config
    ):
        path = config.get('path')
        db = config.get('db')
        if url: 
            if 'http' in scheme and path: url += f'/{path}'
            elif db: url += f'/{db}'    
            return cls(url = url, scheme=scheme, adapter=adapter)
        url = f'{scheme}://'
        host = config.get('host')
        port = config.get('port')
        user = config.get('user')
        password = config.get('password')
        api_key = config.get('api_key')
        token = config.get('token')
        
        if user:
            url += f'{user}'
            if password:
                url += f':{password}'
            url += '@'
        elif api_key:
            url += f'{api_key}@'
        elif token:
            url += f'{token}@'
        url += f'{host}'
        if port: url += f':{port}'
        if 'http' in scheme and path: url += f'/{path}'
        elif db: url += f'/{db}'
        return cls(url = url, scheme = scheme, adapter = adapter)

class BaseDBUrl(UrlModel):

    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'http',
        adapter: Optional[str] = None,
        **config
    ):
        config_key = cls.__name__.split('DB', 1)[0].lower()
        db = config.get(f'{config_key}_db', config.get('db'))
        path = config.get(f'{config_key}_path', config.get('path'))
        
        if url: 
            if 'http' in scheme and path: url += f'/{path}'
            elif db: url += f'/{db}'    
            return cls(url = url, scheme=scheme, adapter=adapter)

        
        url = f'{scheme}://'
        host = config.get(f'{config_key}_host', config.get('host'))
        port = config.get(f'{config_key}_port', config.get('port'))
        user = config.get(f'{config_key}_user', config.get('user'))
        password = config.get(f'{config_key}_password', config.get('password'))
        api_key = config.get(f'{config_key}_api_key', config.get('api_key'))
        token = config.get(f'{config_key}_token', config.get('token'))
        
        if user:
            url += f'{user}'
            if password: url += f':{password}'
            url += '@'
        elif api_key:
            url += f'{api_key}@'
        elif token:
            url += f'{token}@'
        url += f'{host}'
        if port:
            url += f':{port}'
        if 'http' in scheme and path: url += f'/{path}'
        elif db: url += f'/{db}'
        return cls(url = url, scheme = scheme, adapter = adapter)


class PostgresDB(BaseDBUrl):
    url: AnyUrl
    scheme: Optional[str] = 'postgresql'
    adapter: Optional[str] = 'postgresql'
    
    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'postgresql',
        adapter: Optional[str] = 'postgresql',
        **config
    ):
        return super().parse(url = url, scheme = scheme, adapter = adapter, **config)

class RedisDB(BaseDBUrl):
    url: AnyUrl
    scheme: Optional[str] = 'redis'
    adapter: Optional[str] = None
    
    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'redis',
        adapter: Optional[str] = None,
        **config
    ):
        return super().parse(url = url, scheme = scheme, adapter = adapter, **config)

class NatsDB(BaseDBUrl):
    url: AnyUrl
    scheme: Optional[str] = 'nats'
    adapter: Optional[str] = None
    
    @classmethod
    def parse(
        cls, 
        url: Optional[str] = None, 
        scheme: Optional[str] = 'nats',
        adapter: Optional[str] = None,
        **config
    ):
        return super().parse(url = url, scheme = scheme, adapter = adapter, **config)



__all__ = [
    'Field',
    'BaseModel',
    'BaseSettings',
    'Schema',
    'validator',
    'UrlModel',
    'BaseDBUrl',
    'PostgresDB',
    'RedisDB',
    'NatsDB',
]


"""
Proxied Types
"""

from typing import Dict, TypeVar, Optional, Union, Any, TYPE_CHECKING
from .registry import (
    get_global_client, 
    get_client, 
    register_client, 
    update_client_registry_mapping, 
    ClientT, 
    ClientTypeT
)
from lazyops.types.lazydict import LazyDict, RT

if TYPE_CHECKING:
    from aiokeydb import TaskQueue
    from lazyops.types.models import BaseSettings


class ClientProxyMetaClass(type):
    """
    A Custom Proxy that lazily defers initialization of a client until it is called
    """

    _clients: Dict[str, Union[ClientT, ClientTypeT]] = {}
    _module_name: str = None
    is_global: Optional[bool] = True

    @property
    def module_name(cls) -> str:
        """
        Returns the module name
        """
        if cls._module_name is None:
            cls._module_name = cls.__module__.split('.', 1)[0].strip()
        return cls._module_name

    def set_client_mapping(cls, mapping: Dict[str, str]):
        """
        Sets the client mapping
        """
        for key in mapping:
            if cls.module_name not in key:
                mapping[f'{cls.module_name}.{key}'] = mapping.pop(key)
        update_client_registry_mapping(mapping = mapping)

    def add_client_mapping(cls, name: str, module_path: str):
        """
        Adds a client mapping
        """
        if cls.module_name not in name:
            name = f'{cls.module_name}.{name}'
        update_client_registry_mapping(mapping = {name: module_path})

    def get_or_init(cls, name: str, **kwargs) -> Union[ClientT, ClientTypeT]:
        """
        Initializes the client if it is not already initialized
        """
        if name not in cls._clients:
            key = f'{cls.module_name}.{name}' if cls.module_name not in name else name
            if cls.is_global:
                cls._clients[name] = get_global_client(name = key)
            else:
                cls._clients[name] = get_client(name = key, **kwargs)
        return cls._clients[name]
    
    def register(cls, client: Union[ClientT, ClientTypeT], name: str):
        """
        Registers a client
        """
        key = f'{cls.module_name}.{name}' if cls.module_name not in name else name
        if register_client(client = client, name = key) or name not in cls._clients:
            cls._clients[name] = client
    
    def __getitem__(cls, name: str) -> Union[ClientT, ClientTypeT]:
        """
        Returns a client
        """
        return cls.get_or_init(name = name)
    
    def __len__(cls) -> int:
        """
        Returns the number of clients
        """
        return len(cls._clients)
    
    

class QueueManagerMetaClass(type):
    """
    Global Queue Manager
    """
    name: Optional[str] = 'queuemanager'
    _queues: Dict[str, 'TaskQueue'] = {}
    _queue_schemas: Dict[str, Dict[str, Dict[str, Any]]] = None
    
    _settings: Optional['BaseSettings'] = None
    _module_name: str = None

    @property
    def module_name(cls) -> str:
        """
        Returns the module name
        """
        if cls._module_name is None:
            cls._module_name = cls.__module__.split('.', 1)[0].strip()
        return cls._module_name

    def get_settings(cls) -> 'BaseSettings':
        """
        Returns the settings object

        - Should be overwritten by the subclass
        """
        raise NotImplementedError
    
    @property
    def settings(cls) -> 'BaseSettings':
        """
        Returns the settings object
        """
        if cls._settings is None:
            cls._settings = cls.get_settings()
            register_client(cls, name = f'{cls.module_name}.{cls.name}')
        return cls._settings
    
    def get_queue_schemas(cls) -> Dict[str, Dict[str, Dict[str, Any]]]:
        """
        Returns the queue schemas

        - Should be overwritten by the subclass
        """
        raise NotImplementedError
    
    @property
    def queue_schemas(cls) -> Dict[str, Dict[str, Dict[str, Any]]]:
        """
        Returns the queue schemas
        """
        if cls._queue_schemas is None:
            cls._queue_schemas = cls.get_queue_schemas()
        return cls._queue_schemas
    
    def get_task_queue(cls, name: str, kind: Optional[str] = None) -> 'TaskQueue':
        """
        Returns a Task Queue

        - Should be overwritten by the subclass
        """
        raise NotImplementedError
    
    def get_or_init_queue(cls, queue_name: str, kind: Optional[str] = None) -> 'TaskQueue':
        """
        Returns the queue object
        """
        if queue_name not in cls._queues:
            if kind is None:
                for k, v in cls.queue_schemas.items():
                    if queue_name in v:
                        kind = k
                        break
            try:
                q = cls.get_task_queue(
                    name = queue_name,
                    kind = kind,
                )
            except Exception as e:
                q = False
            cls._queues[queue_name] = q
        return cls._queues[queue_name]
    
    def get_queue(cls, queue_name: str, kind: Optional[str] = None) -> Optional['TaskQueue']:
        """
        Returns the queue object
        """
        return cls.get_or_init_queue(queue_name = queue_name, kind = kind)
    

    def __getitem__(cls, queue_name: str) -> Optional['TaskQueue']:
        """
        Returns the queue object
        """
        return cls.get_queue(queue_name = queue_name)
    

class ProxyDict(LazyDict):
    """
    A Proxy Dictionary that lazily defers initialization of a client until it is called

    These are non-global clients
    """
    initialize_objects: Optional[bool] = True
    exclude_schema_attrs: Optional[bool] = True
    proxy_schema: Optional[Dict[str, str]] = None

    def __init__(self, **kwargs):
        if self.proxy_schema is None:
            raise NotImplementedError('Proxy Schema not set')
        self._dict = self.proxy_schema
        if self.exclude_schema_attrs:
            self.excluded_attrs = list(self.proxy_schema.keys())
        self.post_init(**kwargs)
    
    def post_init(self, **kwargs):
        """
        Post Initialization to be overwritten by the subclass
        """
        pass

    def obj_initializer(self, name: str, obj: RT, **kwargs) -> RT:
        """
        Returns the object initializer

        - Can be overwritten by the subclass to modify the object initialization
        """
        return obj(**kwargs)
    
    def get_or_init(self, name: str, default: Optional[RT] = None) -> RT:
        """
        Get an attribute from the dictionary
        If it does not exist, set it to the default value
        """
        if name not in self._dict:
            if default: self._dict[name] = default
            else: raise ValueError(f"Default value for {name} is None")
        from lazyops.utils.lazy import lazy_import
        if isinstance(self._dict[name], str):
            self._dict[name] = lazy_import(self._dict[name])
            if self.initialize_objects:
                self._dict[name] = self.obj_initializer(name, self._dict[name])
        elif isinstance(self._dict[name], tuple):
            obj_class, kwargs = self._dict[name]
            if isinstance(obj_class, str):
                obj_class = lazy_import(obj_class)
            for k,v in kwargs.items():
                if callable(v): kwargs[k] = v()
            self._dict[name] = self.obj_initializer(name, obj_class, **kwargs)
        elif isinstance(self._dict[name], dict):
            self._dict[name] = self.obj_initializer(name, type(self), **self._dict[name])
        return self._dict[name]



"""
GZIP Compress
"""

import gzip
from .base import BaseCompress
from typing import Union

class GzipCompress(BaseCompress):

    @classmethod
    def compress(cls, data: Union[str, bytes], encoding: str = 'utf-8', **kwargs) -> bytes:
        """
        GZip Compress
        """
        if isinstance(data, str): data = data.encode(encoding)
        return gzip.compress(data)

    @classmethod
    def decompress(cls, data: Union[str, bytes], encoding: str = 'utf-8', **kwargs) -> bytes:
        """
        GZip Decompress
        """
        if isinstance(data, str): data = data.encode(encoding = encoding)
        return gzip.decompress(data)




import os

os.environ['POSTGRES_URI'] = 'test:test@localhost:5432/test'
os.environ['POSTGRES_USER'] = 'postgres'
# os.environ['POSTGRES_PASSWORD'] = 'test'

import asyncio
from lazyops.libs.psqldb import PostgresDB

async def test():
    PostgresDB.prepare_db(
        role_user = 'test',
        role_password = 'test',
        db_name = 'test',
    )
    async with PostgresDB.async_session() as sess:
        await sess.execute("SELECT 1")


asyncio.run(test())